@@ -20,6 +20,7 @@
 #if defined(LOG_TO_SYSLOG) || defined(LOG_FAULTS_TO_SYSLOG)
 #include <syslog.h>
 #endif
+#include <paths.h>
 #include <signal.h>
 #include <fcntl.h>
 #include <pwd.h>
@@ -89,14 +90,19 @@ static int	get_site_from_headers(int sock, char *hostname, int hostlen,
 								  char *headers, int header_max,
 								  int *header_len);
 static void alarm_signal(int sig);
+static void log_time(void);
 static void append_string(char *dest, int max_size, char *source);
 static void append_char(char *dest, int max_size, char source);
 
 /*
  * Command line switches.
  */
+static int				daemonize = 1;
 static int				fully_transparent = 0;
 static char				*force_url = NULL;
+#ifdef LOG_TO_FILE
+static FILE				*log_file = NULL;
+#endif
 
 #ifdef IPFILTER
 /*
@@ -118,6 +124,9 @@ int main(int argc, char **argv)
 	short				server_port = -1;
 	uid_t				run_uid = -1;
 	gid_t				run_gid = -1;
+#ifndef DEBUG
+	int					fd;
+#endif
 	int					sock;
 	struct sockaddr_in	addr;
 	int					len;
@@ -125,10 +134,14 @@ int main(int argc, char **argv)
 	/*
 	 * Parse the command line arguments.
 	 */
-	while ((arg = getopt(argc, argv, "ts:r:b:f:")) != EOF)
+	while ((arg = getopt(argc, argv, "dts:r:b:f:l:")) != EOF)
 	{
 		switch (arg)
 		{
+		case 'd':
+			daemonize = 0;
+			break;
+
 		case 't':
 			fully_transparent = 1;
 			break;
@@ -151,6 +164,12 @@ int main(int argc, char **argv)
 			force_url = optarg;
 			break;
 
+		case 'l':
+#ifdef LOG_TO_FILE
+			log_file = fopen(optarg, "a");
+			break;
+#endif
+
 		case '?':
 			usage(argv[0], NULL);
 			break;
@@ -196,6 +215,13 @@ int main(int argc, char **argv)
 		usage(argv[0], "No server port was specified (or it was invalid).");
 	}
 
+#ifdef PATH_LOG_FILE
+	if (log_file == NULL)
+	{
+		log_file = fopen(PATH_LOG_FILE, "a");
+	}
+#endif
+
 #ifdef IPFILTER
 	/*
 	 * Now is a good time to open /dev/ipnat, before giving up our uid/gid.
@@ -218,69 +244,92 @@ int main(int argc, char **argv)
 		sock = bind_to_port(bind_ip, bind_port);
 
 		/*
-		 * Start a server process. When DEBUG is defined, just run
-		 * in the foreground.
+		 * Start a server process.
 		 */
-#ifdef DEBUG
-		switch (0)
-#else
-		switch (fork())
-#endif
+		if (daemonize)
 		{
-		case -1:
-			perror("fork()");
-			exit(1);
-
-		case 0:
-#if defined(LOG_TO_SYSLOG) || defined(LOG_FAULTS_TO_SYSLOG)
-			/*
-			 * Open syslog for logging errors.
-			 */
-			openlog(argv[0], LOG_PID, LOG_DAEMON);
-#endif
-
-			/*
-			 * Ignore some signals.
-			 */
-			signal(SIGHUP, SIG_IGN);
-#ifndef DEBUG
-			signal(SIGINT, SIG_IGN);
-#endif
-			signal(SIGQUIT, SIG_IGN);
-			signal(SIGTSTP, SIG_IGN);
-			signal(SIGCONT, SIG_IGN);
-			signal(SIGPIPE, SIG_IGN);
-			signal(SIGALRM, alarm_signal);
-
-			/*
-			 * Drop back to an untrusted user.
-			 */
-			setgid(run_gid);
-			setuid(run_uid);
+			switch (fork())
+			{
+			case -1:
+				/*
+				 * We could not fork a daemon process.
+				 */
+				perror("fork()");
+				exit(1);
+			case 0:
+				/*
+				 * The child continues as the daemon process.
+				 */
+				break;
+			default:
+				/*
+				 * Parent exits at this stage.
+				 */
+				_exit(0);
+			}
 
 			/*
 			 * Start a new session and group.
 			 */
 			setsid();
-
-			/*
-			 * Handle the server main loop.
-			 */
-			server_main_loop(sock, server_hostname, server_port);
-
-			/*
-			 * Should never exit.
-			 */
-#if defined(LOG_TO_SYSLOG) || defined(LOG_FAULTS_TO_SYSLOG)
-			closelog();
-#endif
-			exit(1);
 		}
 
 		/*
-		 * Parent exits at this stage.
+		 * Set the current directory to the root filesystem.
 		 */
-		exit(0);
+		chdir("/");
+
+#ifndef DEBUG
+		/*
+		 * Point file descriptors to /dev/null, unless DEBUG is defined.
+		 */
+		if ((fd = open(_PATH_DEVNULL, O_RDWR, 0)) >= 0)
+		{
+			dup2(fd, STDIN_FILENO);
+			dup2(fd, STDOUT_FILENO);
+			dup2(fd, STDERR_FILENO);
+			if (fd > STDERR_FILENO)
+				close(fd);
+		}
+#endif
+
+#if defined(LOG_TO_SYSLOG) || defined(LOG_FAULTS_TO_SYSLOG)
+		/*
+		 * Open syslog for logging errors.
+		 */
+		openlog(argv[0], LOG_PID, LOG_DAEMON);
+#endif
+
+		/*
+		 * Ignore some signals.
+		 */
+		signal(SIGHUP, SIG_IGN);
+		if (daemonize)
+			signal(SIGINT, SIG_IGN);
+		signal(SIGQUIT, SIG_IGN);
+		signal(SIGTSTP, SIG_IGN);
+		signal(SIGCONT, SIG_IGN);
+		signal(SIGPIPE, SIG_IGN);
+		signal(SIGALRM, alarm_signal);
+
+		/*
+		 * Drop back to an untrusted user.
+		 */
+		setgid(run_gid);
+		setuid(run_uid);
+
+		/*
+		 * Handle the server main loop.
+		 */
+		server_main_loop(sock, server_hostname, server_port);
+
+		/*
+		 * Should never exit.
+		 */
+#if defined(LOG_TO_SYSLOG) || defined(LOG_FAULTS_TO_SYSLOG)
+		closelog();
+#endif
+		exit(1);
 	}
 
 #if defined(LOG_TO_SYSLOG) || defined(LOG_FAULTS_TO_SYSLOG)
@@ -335,12 +384,20 @@ static void usage(char *prog, char *opt)
 	{
 		fprintf(stderr, "%s: %s\n", prog, opt);
 	}
-	fprintf(stderr, "usage: %s [-t] [-f url] [-s port [-r user] [-b ipaddr]] proxyhost proxyport\n", prog);
+#ifdef LOG_TO_FILE
+	fprintf(stderr, "usage: %s [-dt] [-f url] [-s port [-r user] [-b ipaddr]] [-l file] proxyhost proxyport\n", prog);
+#else
+	fprintf(stderr, "usage: %s [-dt] [-f url] [-s port [-r user] [-b ipaddr]] proxyhost proxyport\n", prog);
+#endif
+	fprintf(stderr, "    -d          When running as a server, do not background the daemon.\n");
 	fprintf(stderr, "    -t          Act fully transparently (default is HTTP translate).\n");
 	fprintf(stderr, "    -f url      Force access to always go to specified URL.\n");
 	fprintf(stderr, "    -s port     Run as a server bound to the specified port.\n");
 	fprintf(stderr, "    -r user     Run as the specified user in server mode.\n");
 	fprintf(stderr, "    -b ipaddr   Bind to the specified ipaddr in server mode.\n");
+#ifdef LOG_TO_FILE
+	fprintf(stderr, "    -l file     Log accesses to the specified file.\n");
+#endif
 	exit(1);
 }
 
@@ -806,6 +863,13 @@ static void trans_proxy(int sock, struct sockaddr_in *from_addr,
 		 */
 #ifdef LOG_TO_SYSLOG
 		syslog(LOG_INFO, "Transparent %s -> %s", from_host, to_host);
+#endif
+#ifdef LOG_TO_FILE
+		if (log_file)
+		{
+			log_time();
+			fprintf(log_file, "Transparent %s -> %s\n", from_host, to_host);
+		}
 #endif
 	}
 	else
@@ -828,6 +892,13 @@ static void trans_proxy(int sock, struct sockaddr_in *from_addr,
 			 */
 #ifdef LOG_TO_SYSLOG
 			syslog(LOG_INFO, "Request NoDNS %s -> %s", from_host, to_host);
+#endif
+#ifdef LOG_TO_FILE
+			if (log_file)
+			{
+				log_time();
+				fprintf(log_file, "Request NoDNS %s -> %s\n", from_host, to_host);
+			}
 #endif
 			break;
 
@@ -842,6 +913,13 @@ static void trans_proxy(int sock, struct sockaddr_in *from_addr,
 			 */
 #ifdef LOG_TO_SYSLOG
 			syslog(LOG_INFO, "Request %s -> %s", from_host, to_host);
+#endif
+#ifdef LOG_TO_FILE
+			if (log_file)
+			{
+				log_time();
+				fprintf(log_file, "Request %s -> %s\n", from_host, to_host);
+			}
 #endif
 			break;
 
@@ -857,6 +935,13 @@ static void trans_proxy(int sock, struct sockaddr_in *from_addr,
 #ifdef LOG_TO_SYSLOG
 		syslog(LOG_NOTICE, "Request In: '%s'", headers);
 #endif
+#ifdef LOG_TO_FILE
+		if (log_file)
+		{
+			log_time();
+			fprintf(log_file, "Request In: '%s'\n", headers);
+		}
+#endif
 
 #ifdef DEBUG
 		fprintf(stderr, "Request In: '%s'\n", headers);
@@ -972,6 +1057,13 @@ static void trans_proxy(int sock, struct sockaddr_in *from_addr,
 #ifdef LOG_TO_SYSLOG
 		syslog(LOG_NOTICE, "Request Out: '%s'", proxy_request);
 #endif
+#ifdef LOG_TO_FILE
+		if (log_file)
+		{
+			log_time();
+			fprintf(log_file, "Request Out: '%s'\n", proxy_request);
+		}
+#endif
 
 #ifdef DEBUG
 		fprintf(stderr, "Request Out: '%s'\n", proxy_request);
@@ -987,6 +1079,13 @@ static void trans_proxy(int sock, struct sockaddr_in *from_addr,
 #ifdef LOG_TO_SYSLOG
 			syslog(LOG_NOTICE, "Forced '%s'", &proxy_request[url_start]);
 #endif
+#ifdef LOG_TO_FILE
+			if (log_file)
+			{
+				log_time();
+				fprintf(log_file, "Forced '%s'\n", &proxy_request[url_start]);
+			}
+#endif
 #ifdef DEBUG
 			fprintf(stderr, "Forced '%s'\n", &proxy_request[url_start]);
 #endif
@@ -996,6 +1095,13 @@ static void trans_proxy(int sock, struct sockaddr_in *from_addr,
 #ifdef LOG_TO_SYSLOG
 			syslog(LOG_NOTICE, "Translated '%s'", &proxy_request[url_start]);
 #endif
+#ifdef LOG_TO_FILE
+			if (log_file)
+			{
+				log_time();
+				fprintf(log_file, "Translated '%s'\n", &proxy_request[url_start]);
+			}
+#endif
 #ifdef DEBUG
 			fprintf(stderr, "Translated '%s'\n", &proxy_request[url_start]);
 #endif
@@ -1005,6 +1111,13 @@ static void trans_proxy(int sock, struct sockaddr_in *from_addr,
 #ifdef LOG_TO_SYSLOG
 			syslog(LOG_NOTICE, "Untranslated '%s'", &proxy_request[url_start]);
 #endif
+#ifdef LOG_TO_FILE
+			if (log_file)
+			{
+				log_time();
+				fprintf(log_file, "Untranslated '%s'\n", &proxy_request[url_start]);
+			}
+#endif
 #ifdef DEBUG
 			fprintf(stderr, "Untranslated '%s'\n", &proxy_request[url_start]);
 #endif
@@ -1025,9 +1138,9 @@ static void trans_proxy(int sock, struct sockaddr_in *from_addr,
 		max_fd = FD_MAX(sock, proxy);
 
 		/*
-		 * Allow 5 minutes for any activity.
+		 * Allow 2 minutes for any activity.
 		 */
-		alarm(300);
+		alarm(120);
 
 		/*
 		 * Wait for some data to be read.
@@ -1209,10 +1322,30 @@ static void alarm_signal(int sig)
 {
 #if defined(LOG_TO_SYSLOG) || defined(LOG_FAULTS_TO_SYSLOG)
 	syslog(LOG_DEBUG, "Alarm signal caught - connection timeout");
+#endif
+#ifdef LOG_TO_FILE
+	if (log_file)
+	{
+		log_time();
+		fprintf(log_file, "Alarm signal caught - connection timeout\n");
+	}
 #endif
 	exit(1);
 }
 
+static void log_time(void)
+{
+	time_t	now;
+	char	date[64];
+
+	if (log_file)
+	{
+		time(&now);
+		strftime(date, sizeof(date), "%a, %d %b %Y %H:%M:%S", localtime(&now));
+		fprintf(log_file, "%s  ", date);
+	}
+}
+
 /*
  * Appends source string to destination string.
  */
